from GTGT.variant_validator import lookup_variant
from .provider import Provider
from .wrappers import lookup_transcript
from flask import Flask, render_template

from typing import Optional

app = Flask(__name__)
provider = Provider()


@app.route("/")
@app.route("/<variant>")
def result(variant: Optional[str] = None) -> str:
    template_file = "index.html.j2"
    if not variant:
        return render_template(template_file)

    # Analyze the transcript
    transcript_id = variant.split(":")[0]
    transcript_model = lookup_transcript(provider, transcript_id)
    transcript = transcript_model.to_transcript()
    results = transcript.analyze(variant)

    # Get external links
    links = lookup_variant(provider, variant).url_dict()

    return render_template(template_file, results=results, links=links, variant=variant)



from copy import deepcopy

from GTGT.mutalyzer import mutation_to_cds_effect, HGVS, exonskip

from .bed import Bed

from typing import List, Dict

TranscriptComparison = Dict[str, float]


class Transcript:
    def __init__(self, exons: Bed, cds: Bed):
        self.exons = exons
        self.cds = cds

        # Set the coding region
        coding = deepcopy(self.exons)
        coding.name = "coding"
        coding.intersect(self.cds)
        self.coding = coding

    def records(self) -> List[Bed]:
        """Return the Bed records that make up the Transcript"""
        return [self.exons, self.cds, self.coding]

    def intersect(self, selector: Bed) -> None:
        """Update transcript to only contain features that intersect the selector"""
        for record in self.records():
            record.intersect(selector)

    def overlap(self, selector: Bed) -> None:
        """Update transcript to only contain features that overlap the selector"""
        for record in self.records():
            record.overlap(selector)

    def subtract(self, selector: Bed) -> None:
        """Remove all features from transcript that intersect the selector"""
        for record in self.records():
            record.subtract(selector)

    def exon_skip(self, selector: Bed) -> None:
        """Remove the exon(s) that overlap the selector from the transcript"""
        exons_to_skip = deepcopy(self.exons)
        exons_to_skip.overlap(selector)
        self.subtract(exons_to_skip)

    def compare(self, other: object) -> Dict[str, float]:
        """Compare the size of each record in the transcripts"""
        if not isinstance(other, Transcript):
            raise NotImplementedError

        # Compare each record that makes up self and other
        # The comparison will fail if the record.name does not match
        cmp = dict()
        for record1, record2 in zip(self.records(), other.records()):
            cmp[record1.name] = record1.compare(record2)

        return cmp

    def compare_score(self, other: object) -> float:
        """Compare the size of each records in the transcripts

        Returns the average value for all records
        """
        if not isinstance(other, Transcript):
            raise NotImplementedError
        cmp = self.compare(other)

        return sum(cmp.values()) / len(cmp)

    def mutate(self, hgvs: str) -> None:
        """Mutate the transcript based on the specified hgvs description"""
        # Determine the CDS interval that is affected by the hgvs description
        H = HGVS(description=hgvs)
        chromStart, chromEnd = mutation_to_cds_effect(H)
        # Subtract that region from the annotations
        self.subtract(
            Bed(chrom=self.cds.chrom, chromStart=chromStart, chromEnd=chromEnd)
        )

    def analyze(self, hgvs: str) -> Dict[str, TranscriptComparison]:
        """Analyse the transcript based on the specified hgvs description

        Calculate the score for the Wildtype (1), the patient transcript and the exon skips
        """
        # Initialize the results dictionary. Wildtype has a score of 1 by definition
        results = dict()
        results["wildtype"] = self.compare(self)

        # Determine the score of the patient
        patient = deepcopy(self)
        patient.mutate(hgvs)
        results["patient"] = patient.compare(self)

        # Determine the score of each exon skip
        for skip in exonskip(HGVS(description=hgvs)):
            # Add deletion to the patient mutation
            desc = HGVS(description=hgvs)
            desc.apply_deletion(skip)

            # Apply the combination to the wildtype transcript
            therapy = deepcopy(self)
            try:
                therapy.mutate(desc.description)
            # Splice site error from mutalyzer, no protein prediction
            except KeyError:
                continue
            results[skip.description] = therapy.compare(self)

        return results



from typing import Any, Dict, Optional
import json
import urllib.request
from urllib.error import HTTPError
from urllib.parse import urlparse
import os
import logging

payload = Dict[str, Any]

logger = logging.getLogger(__name__)


class Provider:
    def __init__(self, cache_dir: Optional[str] = None):
        self.cache_dir = cache_dir

    def get(self, url: str) -> payload:
        if self.cache_dir is None:
            return self._fetch_url(url)
        else:
            return self._fetch_cache(url)

    def _fetch_url(self, url: str) -> payload:
        logger.debug(f"Fetching {url=}")
        try:
            response = urllib.request.urlopen(url)
        except HTTPError as e:
            raise RuntimeError(e)

        data = response.read()

        try:
            js: Dict[str, Any] = json.loads(data)
        except Exception as e:
            logger.error(data)
            raise e

        return js

    def _fetch_cache(self, url: str) -> payload:
        """Fetch cache based on the provided url

        If no cache exists, use self._fetch_url to get and store the data
        """
        fname = f"{self.cache_dir}/{self.url_to_filename(url)}"
        data: payload = dict()

        # Try to open the cache file
        if os.path.exists(fname):
            with open(fname) as fin:
                data = json.load(fin)
                logger.debug(f"Retrieved cache for {url=}")
        else:
            data = self._fetch_url(url)

            # Ensure the folder exists
            os.makedirs(self.cache_dir, exist_ok=True)  # type: ignore

            with open(fname, "w") as fout:
                logger.debug(f"Writing cache for {url=}")
                print(json.dumps(data, indent=True), file=fout)
        return data

    def url_to_filename(self, url: str) -> str:
        fname = ""
        parsed = urlparse(url)
        fname += parsed.netloc
        if len(parsed.path) > 1:
            fname += parsed.path.replace("/", "_")

        if parsed.query:
            fname += "_" + parsed.query.replace("/", "_")

        return fname



from copy import deepcopy
from mutalyzer.description import Description, to_rna_reference_model, model_to_string
from mutalyzer.converter.to_hgvs_coordinates import to_hgvs_locations
from mutalyzer.converter.to_internal_coordinates import to_internal_coordinates
from mutalyzer.converter.to_internal_indexing import to_internal_indexing
from mutalyzer.checker import is_overlap
import mutalyzer_hgvs_parser

from pydantic import BaseModel, model_validator

from typing import Any, Tuple, List, Dict

VariantModel = Dict[str, Any]


class HGVS(BaseModel):
    description: str

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "description": "NM_002520.7:c.860_863dup",
                }
            ]
        }
    }

    @model_validator(mode="after")
    def hgvs_parser(self) -> "HGVS":
        """Parse the HGVS description with mutalyzer-hgvs-parser"""
        hgvs_error = (
            mutalyzer_hgvs_parser.exceptions.UnexpectedCharacter,
            mutalyzer_hgvs_parser.exceptions.UnexpectedEnd,
        )
        try:
            mutalyzer_hgvs_parser.to_model(self.description)
        except hgvs_error as e:
            raise ValueError(e)
        return self

    @staticmethod
    def _validate_for_apply_deletion(hgvs: "HGVS") -> None:
        """
        Raise a NotImplementedError if the hgvs is not supported
        """
        model = mutalyzer_hgvs_parser.to_model(hgvs.description)

        # There must be one variant
        if len(model["variants"]) != 1:
            raise NotImplementedError

        var = model["variants"][0]

        # The variant must be in the CDS
        if "outside_cds" in var["location"]:
            raise NotImplementedError

        # The variant must not be intronic
        if "offset" in var["location"]:
            raise NotImplementedError

    @property
    def position(self) -> Tuple[int, int]:
        """
        Return the position of a description as (start, end)

        These are just the .c position, so 1 based and inclusive
        """
        model = mutalyzer_hgvs_parser.to_model(self.description)
        assert len(model["variants"]) == 1

        var = model["variants"][0]
        if var["location"]["type"] == "point":
            p = var["location"]["position"]
            return p, p
        elif var["location"]["type"] == "range":
            s = var["location"]["start"]["position"]
            e = var["location"]["end"]["position"]
            return s, e
        else:
            raise NotImplementedError

    def apply_deletion(self, other: "HGVS") -> None:
        """
        Apply a deletion to the current variant

        If the deletion does not overlap, add them together
        If the deletion completely overlaps the variant, replace the variant

        If the deletion partially overlaps the variant, raise an error
        """
        # Perform all validations
        self._validate_for_apply_deletion(other)
        self._validate_for_apply_deletion(self)

        # other must be a deletion
        o_model = mutalyzer_hgvs_parser.to_model(other.description)
        assert len(o_model["variants"]) == 1

        o_type = o_model["variants"][0]["type"]
        if o_type != "deletion":
            raise NotImplementedError

        s_model = mutalyzer_hgvs_parser.to_model(self.description)
        assert len(s_model["variants"]) == 1
        s_type = s_model["variants"][0]["type"]

        # self and other must refer to the same reference ID
        s_id = s_model["reference"]["id"]
        o_id = o_model["reference"]["id"]
        if s_id != o_id:
            raise NotImplementedError

        # Get the c. positions for start and end
        s_start, s_end = self.position
        o_start, o_end = other.position

        # Get the variants in text format
        s_var = self.description.split("c.")[1]
        o_var = other.description.split("c.")[1]

        # If self is a deletion, and other is fully inside self, we don't have to add anything
        if s_type == "deletion":
            if o_start >= s_start and o_end <= s_end:
                return

        # If self is not a deletion:
        # self is before other
        if s_end < o_start:
            self.description = f"{s_id}:c.[{s_var};{o_var}]"
        # self is after other
        elif s_start > o_end:
            self.description = f"{s_id}:c.[{o_var};{s_var}]"
        # self is fully inside other
        elif s_start >= o_start and s_end <= o_end:
            # We overwrite self with other
            self.description = other.description
        # partial overlaps are not supported
        else:
            msg = f"Unable to apply deletion {other} to {self}"
            raise NotImplementedError(msg)


def HGVS_to_genome_range(hgvs: HGVS) -> Tuple[int, int]:
    """Convert HGVS variant description to affected genome range

    NOTE that the genome range is similar to the UCSC annotations on the genome,
    i.e. 0 based, half open. Not to be confused with hgvs g. positions
    """
    d = Description(description=hgvs.description)
    d.normalize()
    model = d.ensembl_model_with_no_offset()

    if len(model["variants"]) == 0:
        raise ValueError("Descriptions without variants are not supported")
    if len(model["variants"]) > 1:
        raise ValueError("Multiple variants are not supported")

    # Get start and end of the description
    start = model["variants"][0]["location"]["start"]["position"]
    end = model["variants"][0]["location"]["end"]["position"]

    return (start, end)


def exonskip(hgvs: HGVS) -> List[HGVS]:
    """Generate all possible exon skips for the specified HGVS description"""
    d = Description(description=hgvs.description)
    d.to_delins()

    # Extract relevant information from the normalized description
    raw_response = d.output()
    exons = raw_response["selector_short"]["exon"]["c"]
    transcript_id = raw_response["input_model"]["reference"]["id"]

    exon_skips = list()
    # The first and second exon cannot be skipped
    for start, end in exons[1:-1]:
        skip = f"{transcript_id}:c.{start}_{end}del"
        exon_skips.append(HGVS(description=skip))
    return exon_skips


def _init_model(d: Description) -> None:
    """
    Initialize the HGVS Description

    Don't normalize the positions
    TODO: check that other sanity checks are still performed
    """
    d.to_delins()
    d.de_hgvs_internal_indexing_model = d.delins_model
    d.construct_de_hgvs_internal_indexing_model()
    d.construct_de_hgvs_coordinates_model()
    d.construct_normalized_description()
    d.construct_protein_description()


def mutation_to_cds_effect(hgvs: HGVS) -> Tuple[int, int]:
    """
    Determine the effect of the specified HGVS description on the CDS, on the genome

    Steps:
    - Use the protein prediction of mutalyzer to determine which protein
      residues are changed
    - Map this back to a deletion in c. positions to determine which protein
      annotations are no longer valid
    - Convert the c. positions to genome coordiinates as used by the UCSC
    NOTE that the genome range is similar to the UCSC annotations on the genome,
    i.e. 0 based, half open. Not to be confused with hgvs g. positions
    """
    d = Description(description=hgvs.description)
    # Do only part of the normalizations from mutalyzer
    _init_model(d)

    # Determine the protein positions that were changed
    protein = d.output()["protein"]
    first = protein["position_first"]
    last = protein["position_last_original"]

    # Convert the changed amino acids into a deletion in HGVS c. format
    transcript_id = hgvs.description.split(":c.")[0]
    start_pos = first * 3
    end_pos = (last * 3) - 1

    cdot = f"{transcript_id}:c.{start_pos}_{end_pos}del"

    return HGVS_to_genome_range(HGVS(description=cdot))


def variant_to_model(variant: str) -> List[VariantModel]:
    """
    Parse the specified variant into a variant model
    """
    results: List[VariantModel]
    if "[" in variant:
        results = mutalyzer_hgvs_parser.to_model(variant, "variants")
    else:
        results = [mutalyzer_hgvs_parser.to_model(variant, "variant")]
    return results


def append_mutation(description: Description, mutation: str) -> None:
    """
    Add mutation to the Description, re-using the Description object
    """
    # Get the variant model in c.
    c_variants = variant_to_model(mutation)

    # Convert the c. variant to i.
    model = deepcopy(description.corrected_model)
    # Add the c_variant to the variant(s) which are already there
    model["variants"] += c_variants
    model = to_internal_coordinates(model, description.references)
    model = to_internal_indexing(model)

    if is_overlap(model["variants"]):
        msg = f"Variant {mutation} overlaps {description.input_description}"
        raise ValueError(msg)

    # Replace the variant in the description
    description.de_hgvs_internal_indexing_model["variants"] = model["variants"]

    # Update the internal description models
    description.construct_de_hgvs_coordinates_model()
    description.construct_normalized_description()
    description.construct_protein_description()



from typing import Tuple, List, Set

Range = Tuple[int, int]


def overlap(a: Range, b: Range) -> bool:
    """Determine if ranges a and b overlap"""
    # A and B overlap wether the intersection is not empty
    if intersect(a, b):
        return True
    return False


def intersect(a: Range, b: Range) -> List[Range]:
    """Determine the intersection between ranges a and b"""
    start = max(a[0], b[0])
    end = min(a[1], b[1])

    if start + 1 > end:
        return list()
    else:
        return [(start, end)]


def subtract(a: List[Range], b: List[Range]) -> List[Range]:
    """
    Subtract the regions in b from a

    Lazy implementation by just putting all numbers into two sets
    """
    A: Set[int] = set()
    for start, end in a:
        A.update(range(start, end))

    B: Set[int] = set()
    for start, end in b:
        B.update(range(start, end))

    return _to_range(A - B)


def _to_range(numbers: Set[int]) -> List[Range]:
    """Convert a set of numbers to a range[start, end)"""
    # Make sure the numbers are sorted
    _numbers = sorted(numbers)

    # If there are no _numbers
    if not _numbers:
        return list()

    # If there is only a single number
    if len(_numbers) == 1:
        i = _numbers[0]
        return [(i, i + 1)]

    # Initialise the start and previous number
    start = prev = _numbers[0]

    # Store the ranges we found
    ranges = list()

    # Process all _numbers
    for i in _numbers[1:]:
        if i == prev + 1:
            prev = i
        else:
            ranges.append((start, prev + 1))
            start = prev = i
    ranges.append((start, prev + 1))

    return ranges



from pydantic import BaseModel, model_validator
from enum import Enum
from typing import Any, Dict, List, Tuple, Union
from .bed import Bed
from .transcript import Transcript

from pydantic import Field

import mutalyzer_hgvs_parser

Range = Tuple[int, int]


class Assembly(Enum):
    HUMAN = "GRCh38"
    RAT = "mRatBN7.2"


class EnsemblTranscript(BaseModel):
    assembly_name: Assembly
    seq_region_name: str
    start: int
    end: int
    id: str
    version: int
    display_name: str


class BedModel(BaseModel):
    """Pydantic wrapper for BED objects"""

    chrom: str
    blocks: List[Range]
    name: str = ""
    score: int = 0
    strand: str = "."

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "chrom": "chr1",
                    "blocks": [[30, 35]],
                    "name": "",
                }
            ]
        }
    }

    @model_validator(mode="after")
    def valid_Bed(self) -> "BedModel":
        """Validate that the data is consistent with Bed requirements"""
        try:
            self.to_bed()
        except ValueError as e:
            raise (e)
        return self

    @classmethod
    def from_ucsc(cls, ucsc: Dict[str, Union[str, int]]) -> "BedModel":
        fields: Dict[str, Any] = ucsc.copy()

        # Note that UCSC calls the "blockStarts" field "chromStarts"
        block_starts = Bed._csv_to_int(fields["chromStarts"])
        block_sizes = Bed._csv_to_int(fields["blockSizes"])

        # Determine the blocks from the payload
        chr_start = fields["chromStart"]
        blocks = list()
        for start, size in zip(block_starts, block_sizes):
            block_start = start + chr_start
            block_end = block_start + size
            blocks.append((block_start, block_end))

        return cls(
            chrom=fields["chrom"],
            blocks=blocks,
            name=fields["name"],
            score=fields["score"],
            strand=fields["strand"],
        )

    def to_bed(self) -> Bed:
        chrom_start = self.blocks[0][0]
        chrom_end = self.blocks[-1][1]
        blockCount = len(self.blocks)
        blockSizes = [end - start for start, end in self.blocks]
        blockStarts = [start - chrom_start for start, _ in self.blocks]

        return Bed(
            chrom=self.chrom,
            chromStart=chrom_start,
            chromEnd=chrom_end,
            name=self.name,
            score=self.score,
            strand=self.strand,
            blockCount=blockCount,
            blockSizes=blockSizes,
            blockStarts=blockStarts,
        )

    @classmethod
    def from_bed(cls, bed: Bed) -> "BedModel":
        return cls(
            chrom=bed.chrom,
            blocks=list(bed.blocks()),
            name=bed.name,
            score=bed.score,
            strand=bed.strand,
        )


class TranscriptModel(BaseModel):
    exons: BedModel
    cds: BedModel

    model_config = {
        "json_schema_extra": {
            "examples": [
                # Example transcript
                {
                    "exons": {
                        "chrom": "chr1",
                        "blocks": [[0, 10], [20, 40], [50, 60], [70, 100]],
                        "name": "exons",
                    },
                    "cds": {
                        "chrom": "chr1",
                        "blocks": [[23, 72]],
                        "name": "cds",
                    },
                },
                # Skipped exon 2 ([20, 40])
                {
                    "exons": {
                        "chrom": "chr1",
                        "blocks": [[0, 10], [50, 60], [70, 100]],
                        "name": "exons",
                        "score": 0,
                        "strand": ".",
                    },
                    "cds": {
                        "chrom": "chr1",
                        "blocks": [[40, 72]],
                        "name": "cds",
                        "score": 0,
                        "strand": ".",
                    },
                },
            ]
        }
    }

    def to_transcript(self) -> Transcript:
        exons = self.exons.to_bed()
        cds = self.cds.to_bed()
        return Transcript(exons=exons, cds=cds)

    @classmethod
    def from_transcript(cls, transcript: Transcript) -> "TranscriptModel":
        """Create a TranscriptModel from a Transcript object"""
        exons = BedModel.from_bed(transcript.exons)
        cds = BedModel.from_bed(transcript.cds)
        return cls(exons=exons, cds=cds)


class TranscriptId(BaseModel):
    id: str = Field(pattern=r"^ENST\d+\.\d+$")

    model_config = {
        "json_schema_extra": {
            "examples": [
                {"id": "ENST00000296930.10"},
            ]
        }
    }



from .models import TranscriptModel, BedModel
from .provider import Provider
from .ensembl import lookup_transcript as lookup_transcript_ens
from .ucsc import lookup_knownGene


def lookup_transcript(provider: Provider, transcript_id: str) -> TranscriptModel:
    r = lookup_transcript_ens(provider, transcript_id)
    # track_name = "ncbiRefSeq"
    # track_name = "ensGene"
    track_name = "knownGene"
    track = lookup_knownGene(provider, r, track_name)
    knownGene = track[track_name][0]
    exons = BedModel.from_ucsc(knownGene)

    # The CDS is defied as the thickStart, thickEnd in ucsc
    chrom = knownGene["chrom"]
    start = knownGene["thickStart"]
    end = knownGene["thickEnd"]
    name = "cds"
    strand = knownGene["strand"]
    cds = BedModel(chrom=chrom, blocks=[(start, end)], name=name, strand=strand)

    return TranscriptModel(exons=exons, cds=cds)



"""
Module that contains the command line app, so we can still import __main__
without executing side effects
"""

from .wrappers import lookup_transcript
from .variant_validator import lookup_variant
from .provider import Provider
from .mutalyzer import exonskip
from .mutalyzer import HGVS

import argparse
import json
import logging
import os


def logger_setup() -> logging.Logger:
    import logging

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    logger.addHandler(ch)

    return logger


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Description of command.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="command")
    parser.add_argument("--cachedir", type=str, default=os.environ.get("GTGT_CACHE"))

    transcript_parser = subparsers.add_parser(
        "transcript", help="Transcript Information"
    )

    transcript_parser.add_argument(
        "transcript_id", type=str, help="Transcript of interest"
    )

    link_parser = subparsers.add_parser("links", help="Links to external resources")
    link_parser.add_argument("hgvs_variant", type=str, help="Variant of interest")

    api_server_parser = subparsers.add_parser(
        "api_server", help="Run the GTGT API server"
    )
    api_server_parser.add_argument(
        "--host", default="localhost", help="Hostname to listen on"
    )

    web_server_parser = subparsers.add_parser(
        "webserver", help="Run the GTGT web server"
    )
    web_server_parser.add_argument(
        "--host", default="localhost", help="Hostname to listen on"
    )
    web_server_parser.add_argument(
        "--debug", default=False, action="store_true", help="Run Flask in debug mode"
    )

    mutator_parser = subparsers.add_parser(
        "mutate", help="Mutate the specified transcript"
    )

    mutator_parser.add_argument("transcript_id", help="The transcript to mutate")

    analyze_parser = subparsers.add_parser(
        "analyze", help="Analyze all possible exons skips for the spcified HGVS variant"
    )
    analyze_parser.add_argument(
        "hgvs", help="HGVS description of the transcript of interest"
    )
    args = parser.parse_args()

    logger = logger_setup()

    provider = Provider(args.cachedir)

    if args.command == "transcript":
        ts = lookup_transcript(provider, args.transcript_id)
        print(ts.model_dump_json())
    elif args.command == "links":
        logger.debug(args)
        links = lookup_variant(provider, args.hgvs_variant).url_dict()
        for website, url in links.items():
            print(f"{website}: {url}")
    elif args.command == "api_server":
        try:
            from .app import app, uvicorn
        except ModuleNotFoundError:
            print("Missing modules, please install with 'pip install gtgt[api_server]'")
            exit(-1)
        uvicorn.run(app, host=args.host)
    elif args.command == "mutate":
        desc = f"{args.transcript_id}:c.="
        hgvs_transcript = HGVS(description=desc)
        for hgvs in exonskip(hgvs_transcript):
            print(hgvs.description)
        print(args.transcript_id)
    elif args.command == "analyze":
        transcript_id = args.hgvs.split(":")[0]
        transcript_model = lookup_transcript(provider, transcript_id)
        transcript = transcript_model.to_transcript()
        results = transcript.analyze(args.hgvs)
        print(json.dumps(results, indent=True))
    elif args.command == "webserver":
        try:
            from .flask import app as flask_app
        except ModuleNotFoundError:
            print("Missing modules, please install with 'pip install gtgt[webserver]'")
            exit(-1)
        flask_app.run(args.host, debug=args.debug)
    else:
        raise NotImplementedError()


if __name__ == "__main__":
    main()
